from typing import Any, Dict

from pyquery import PyQuery

from ...utils import HTTPClient


path = "/_/example/pyquery"


async def provider() -> Dict[str, Any]:
    """Define a basic example data provider function."""
    async with HTTPClient() as client:
        response = await client.get("https://httpbin.org/html")
    dom = PyQuery(response.text)
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": dom("h1").text(), "description": dom("div > p").text()}],
    }



import threading
import time
from functools import partial, wraps
from typing import Any, Awaitable, Callable, Dict, Tuple, TypeVar, cast


RATELIMIT_PERIOD = 3600

AsyncFn = TypeVar("AsyncFn", bound=Callable[..., Awaitable[Any]])


class RateLimitError(Exception):
    """Exception for function rate limit exceeded."""

    def __init__(self, fn: str, called: int, max_calls: int) -> None:
        """Populate `RateLimitError` error message."""
        super().__init__(f"function {fn} ratelimit exceeded ({called} > {max_calls}).")


def decorator(fn: AsyncFn, *, calls: int, period: float) -> AsyncFn:
    """Rate limit decorator."""
    called = 0
    last_reset_time = time.monotonic()
    lock = threading.RLock()
    max_calls = calls

    @wraps(fn)
    async def wrapper(*args: Tuple[Any, ...], **kwds: Dict[str, Any]) -> Any:
        """Wrap the original async `fn`.

        At most `calls` invocations are allow within `period` seconds.

        Args:
            args: Positional arguments in function parameters.
            kwds: Keyword arguments in function parameters.

        Returns:
            The result of `fn(*args, **kwds)`, if rate limit not exceeded.

        Raises:
            RateLimitError: function rate limit exceeded.
        """
        nonlocal called, last_reset_time, max_calls
        with lock:
            current_time = time.monotonic()
            # reset if the period is over
            if current_time > last_reset_time + period:
                called = 0
                last_reset_time = current_time
            called += 1
            # raise if rate limit exceeded
            if called > max_calls:
                raise RateLimitError(
                    f"{fn.__module__}.{fn.__name__}", called, max_calls
                )
            return await fn(*args, **kwds)

    return cast(AsyncFn, wrapper)


def ratelimit(
    *, calls: int, period: float = RATELIMIT_PERIOD
) -> Callable[[AsyncFn], AsyncFn]:
    """Rate limit function calls.

    Args:
        calls: number of function calls allow within one period.
        period: length of the period, in seconds.

    Returns:
        A decorator that accepts an async function and returns a rate limited one.
    """
    return partial(decorator, calls=calls, period=period)



from typing import Any

import httpx
from fake_useragent import UserAgent  # type: ignore[import]
from playwright.async_api import async_playwright
from playwright.async_api._generated import Page


ua = UserAgent(
    fallback="Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 "
    "(KHTML, like Gecko) Chrome/94.0.4606.71 Safari/537.36"
)


class Browser:
    """Wrap playwright browser as a context manager."""

    async def __aenter__(self) -> Page:
        """Enter the context manager."""
        self.context = await async_playwright().start()
        self.browser = await self.context.chromium.launch()
        return await self.browser.new_page()

    async def __aexit__(self, *_: Any) -> None:
        """Leave the context manager."""
        await self.browser.close()
        await self.context.stop()  # type: ignore[func-returns-value]


class HTTPClient(httpx.AsyncClient):
    """Wrap `httpx.AsyncClient` with default headers."""

    def build_request(self, *args: Any, **kwds: Any) -> httpx.Request:
        """Set default Host/Referer/User-Agent for requests."""
        request = super().build_request(*args, **kwds)
        url = request.url
        request.headers.setdefault("Host", url.host)
        request.headers.setdefault("Referer", f"{url.scheme}://{url.host}")
        request.headers.setdefault("User-Agent", ua.random)
        return request



from ...models import Persona, Plugin
from . import (
    example,
    example_cache,
    example_httpx,
    example_playwright,
    example_pyquery,
    example_ratelimit,
    example_with_args,
)


plugin = Plugin(
    name="rsserpent-plugin-builtin",
    author=Persona(
        name="queensferryme",
        link="https://github.com/queensferryme",
        email="queensferry.me@gmail.com",
    ),
    repository="https://github.com/RSSerpent-Rev/RSSerpent",
    prefix="/_",
    routers={
        example.path: example.provider,
        example_cache.path: example_cache.provider,
        example_httpx.path: example_httpx.provider,
        example_playwright.path: example_playwright.provider,
        example_pyquery.path: example_pyquery.provider,
        example_ratelimit.path: example_ratelimit.provider,
        example_with_args.path: example_with_args.provider,
    },
)

__all__ = ("plugin",)



from typing import cast

from importlib_metadata import entry_points

from ..models import Plugin


# https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/#using-package-metadata
plugins = [cast(Plugin, ep.load()) for ep in entry_points(group="rsserpent.plugin")]

__all__ = ("plugins",)



from typing import Any, Dict

from ...utils import cached


count = 0
path = "/_/example/cache"


@cached
async def provider() -> Dict[str, Any]:
    """Define a basic example data provider function with `@cached`."""
    global count
    count += 1
    return {
        "title": f"Example {count}",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": "Example Title", "description": "Example Description"}],
    }



from typing import Any, Dict

from ...utils import ratelimit


path = "/_/example/ratelimit"


@ratelimit(calls=1)
async def provider() -> Dict[str, Any]:
    """Define a basic example data provider function with rate limit."""
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": "Example Title", "description": "Example Description"}],
    }



import os
import threading
import time
from collections import OrderedDict
from dataclasses import dataclass
from functools import partial, wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)


__all__ = ("cached",)

CACHE_EXPIRE = int(os.environ.get("CACHE_EXPIRE", 60 * 60))

AsyncFn = TypeVar("AsyncFn", bound=Callable[..., Awaitable[Any]])


class CacheKey:
    """Hashed function parameters that could be used as dictionary keys.

    This module provides `@cached` & `@cached_with` decorators for caching function
    results. Cache exists as key-value pairs. The parameters of each function
    invocation are converted to a tuple and then hashed, so that results of different
    function invocations (e.g. `fn(1)` & `fn(2)`) could be differentiated and
    separately cached.

    Attributes:
        hashvalue: The hashed value of function parameters.
    """

    __sentinel__ = object()
    __slots__ = "hashvalue"

    def __init__(self, key: Tuple[Any, ...]) -> None:
        self.hashvalue = hash(key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CacheKey):
            raise NotImplementedError()
        return self.hashvalue == other.hashvalue

    def __hash__(self) -> int:
        return self.hashvalue

    @classmethod
    def make(cls, args: Tuple[Any, ...], kwds: Dict[str, Any]) -> "CacheKey":
        """Create a `CacheKey` instance from any function parameters.

        Note that parameters of non-primitive types are discarded.

        Args:
            args: Positional arguments in function parameters.
            kwds: Keyword arguments in function parameters.

        Returns:
            The created `CacheKey` instance.
        """
        key = []
        # positional arguments
        for argument in args:
            key.append(argument)
            key.append(type(argument))
        # keyword arguments
        if kwds:
            key.append(cls.__sentinel__)
            for _, argument in sorted(kwds.items()):
                key.append(argument)
                key.append(type(argument))
        return CacheKey(tuple(key))


@dataclass
class CacheValue:
    """Cached function results with a expiring timestamp."""

    expired: float
    data: Any


class LRUCache(OrderedDict):  # type: ignore[type-arg]
    """Least Recently Used (LRU) cache.

    Attributes:
        hits: The number of cache hits.
        maxsize: The maximum capacity of the LRU cache.
        misses: The number of cache misses.
    """

    def __init__(self, maxsize: int) -> None:
        self.maxsize = maxsize if maxsize > 0 else float("inf")
        self.hits, self.misses = 0, 0
        self.lock = threading.RLock()
        super().__init__()

    def __getitem__(self, key: CacheKey) -> Optional[CacheValue]:
        """Get cache value by key.

        In case of cache miss / expire, `None` is returned. Otherwise the cached value
        is returned and moved to the end of the list.
        """
        if key not in self:
            return None
        value: CacheValue = super().__getitem__(key)
        if value.expired < time.monotonic():
            return None
        self.move_to_end(key)
        self.hits += 1
        return value

    def __setitem__(self, key: CacheKey, value: CacheValue) -> None:
        """Set a cache (key, value) pair.

        The `__setitem__` is called only when a cache miss / expire happens. In case of
        cache expire, the existing `key` must be moved to the end of the list. In case
        cache maximum capacity is reached, the first (thus least recently used) cache
        item is deleted.
        """
        with self.lock:
            if key in self:
                self.move_to_end(key)
            super().__setitem__(key, value)
            self.misses += 1
            if len(self) > self.maxsize:
                # do not use `self.popitem`
                del self[next(iter(self))]

    @property
    def size(self) -> int:
        """Return the current size of the cache."""
        return super().__len__()


def decorator(fn: AsyncFn, *, expire: int, maxsize: int) -> AsyncFn:
    """Cache decorator."""
    cache = LRUCache(maxsize=maxsize)

    @wraps(fn)
    async def wrapper(*args: Tuple[Any, ...], **kwds: Dict[str, Any]) -> Any:
        """Wrap the original async `fn`.

        Cached results will be returned if cache hit, otherwise
        (missing/expired) `fn` will be invoked and its result will be cached.

        Args:
            args: Positional arguments in function parameters.
            kwds: Keyword arguments in function parameters.

        Returns:
            The (maybe cached) result of `fn(*args, **kwds)`.
        """
        key = CacheKey.make(args, kwds)
        value = cache[key]
        # cache miss/expired
        if value is None:
            result = await fn(*args, **kwds)
            cache[key] = CacheValue(expired=time.monotonic() + expire, data=result)
            return result
        return value.data

    wrapper.__dict__["cache"] = cache
    wrapper.__dict__["expire"] = expire
    return cast(AsyncFn, wrapper)


@overload
def cached(fn: AsyncFn) -> AsyncFn:
    ...  # pragma: no cover


@overload
def cached(
    *, expire: int = CACHE_EXPIRE, maxsize: int = 0
) -> Callable[[AsyncFn], AsyncFn]:
    ...  # pragma: no cover


def cached(
    fn: Optional[AsyncFn] = None, *, expire: int = CACHE_EXPIRE, maxsize: int = 0
) -> Union[AsyncFn, Callable[[AsyncFn], AsyncFn]]:
    """Cache function results."""
    if fn is not None:
        return decorator(fn, expire=expire, maxsize=maxsize)
    return partial(decorator, expire=expire, maxsize=maxsize)


def get_cache(fn: AsyncFn) -> Optional[LRUCache]:
    return fn.__dict__.get("cache")



import os
import re
import subprocess
import sys
import traceback
from pathlib import Path

import arrow
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.routing import Route
from starlette.templating import Jinja2Templates, _TemplateResponse as TemplateResponse

from .models import Feed, Plugin, ProviderFn
from .plugins import plugins
from .utils import fetch_data


templates = Jinja2Templates(directory=str(Path(__file__).parent / "templates"))
templates.env.autoescape = True
templates.env.globals["arrow"] = arrow


async def exception_handler(request: Request, exception: Exception) -> TemplateResponse:
    """Return an HTML web page for displaying the current exception."""
    tb = traceback.format_exc()
    for path in sys.path:
        tb, _ = re.subn(rf'(?<=File "){re.escape(path)}[/\\]', "", tb)
    return templates.TemplateResponse(
        "exception.html.jinja",
        {
            "exception": exception,
            "request": request,
            "traceback": tb,
        },
    )


def startup() -> None:
    """Install Chromium on start-up."""
    if sys.platform == "linux" and "PLAYWRIGHT_BROWSERS_PATH" not in os.environ:
        os.environ["PLAYWRIGHT_BROWSERS_PATH"] = "/tmp"  # pragma: is_not_linux
    print("Running on", sys.platform)
    print(f"Installing Chromium at {os.environ.get('PLAYWRIGHT_BROWSERS_PATH')}")
    subprocess.run("playwright install chromium".split())


async def index(request: Request) -> TemplateResponse:
    """Return the index HTML web page."""
    return templates.TemplateResponse("index.html.jinja", {"request": request})


routes = [Route("/", endpoint=index)]


for plugin in plugins:
    for path, provider in plugin.routers.items():

        async def endpoint(
            request: Request, plugin: Plugin = plugin, provider: ProviderFn = provider
        ) -> TemplateResponse:
            """Return an RSS feed of XML format."""
            data = await fetch_data(
                provider, request.path_params, dict(request.query_params)
            )
            return templates.TemplateResponse(
                "rss.xml.jinja",
                {"data": Feed(**data), "plugin": plugin, "request": request},
                media_type="application/xml",
            )

        routes.append(Route(path, endpoint=endpoint))


app = Starlette(
    debug=bool(os.environ.get("DEBUG")),
    exception_handlers={Exception: exception_handler},
    on_startup=[startup],
    routes=routes,
)



from typing import TYPE_CHECKING, Any, Dict, List, Optional

import arrow
from arrow import Arrow
from pydantic import BaseModel, Field, root_validator


if TYPE_CHECKING:
    AnyUrl = str
    HttpUrl = str
else:
    from pydantic import AnyUrl, HttpUrl


class RSSModelError(ValueError):
    """Exception for `Feed` model validation error."""

    empty_title_and_description = (
        "at least one of <title> or <description> must be present."
    )


class Category(BaseModel):
    """Data model for the `<category>` field in an RSS 2.0 feed."""

    name: str
    domain: Optional[HttpUrl]


class Enclosure(BaseModel):
    """Data model for the `<enclosure>` field in an RSS 2.0 feed."""

    length: int = 0
    type: str
    url: AnyUrl


class Guid(BaseModel):
    """Data model for the `<guid>` field in an RSS 2.0 feed."""

    is_perma_link: bool = True
    value: str


class Image(BaseModel):
    """Data model for the `<image>` field in an RSS 2.0 feed."""

    url: HttpUrl
    title: str
    link: HttpUrl
    width: Optional[int] = 88
    height: Optional[int] = 31
    description: Optional[str]


class Source(BaseModel):
    """Data model for the `<source>` field in an RSS 2.0 feed."""

    name: str
    url: Optional[HttpUrl]


class Item(BaseModel):
    """Data model for the `<item>` field in an RSS 2.0 feed."""

    title: Optional[str]
    link: Optional[HttpUrl]
    description: Optional[str]
    author: Optional[str]
    categories: Optional[List[Category]]
    comments: Optional[HttpUrl]
    enclosure: Optional[Enclosure]
    guid: Optional[Guid]
    pub_date: Optional[Arrow]
    source: Optional[Source]

    class Config:  # noqa: D106
        arbitrary_types_allowed = True

    @root_validator
    def validate(  # type: ignore[override]
        cls, values: Dict[str, Any]  # noqa: N805
    ) -> Dict[str, Any]:
        """Ensure at least one of `<title>` or `<description>` is present."""
        title, description = values.get("title"), values.get("description")
        if title is None and description is None:
            raise RSSModelError(RSSModelError.empty_title_and_description)
        return values


class Feed(BaseModel):
    """Data model for RSS 2.0 feeds.

    See specification at https://www.rssboard.org/rss-specification.
    Note that some rarely used fields are not implemented:
    * `<cloud>`
    * `<rating>`
    * `<textInput>`
    * `<skipHours>`
    * `<skipDays>`
    """

    title: str
    link: HttpUrl
    description: str
    language: Optional[str]
    copyright: Optional[str]
    managing_editor: Optional[str]
    web_master: Optional[str]
    pub_date: Optional[Arrow] = Field(default_factory=arrow.utcnow)
    last_build_date: Optional[Arrow] = Field(default_factory=arrow.utcnow)
    categories: Optional[List[Category]]
    generator: Optional[str] = __package__.split(".")[0]
    docs: Optional[HttpUrl] = "https://www.rssboard.org/rss-specification"
    ttl: Optional[int] = 60
    image: Optional[Image]
    items: Optional[List[Item]]

    class Config:  # noqa: D106
        arbitrary_types_allowed = True



from typing import Any, Dict

from ...utils import HTTPClient


path = "/_/example/httpx"


async def provider() -> Dict[str, Any]:
    """Define a basic example data provider function."""
    async with HTTPClient() as client:
        response = await client.get("https://httpbin.org/ip")
        ip = response.json()["origin"]
    return {
        "title": f"{ip}",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": "Example Title", "description": "Example Description"}],
    }



from .builtin import plugin as builtin_plugin
from .externals import plugins as external_plugins


plugins = [builtin_plugin, *external_plugins]

for p in plugins:
    print(f"Loaded plugin: {p.name}, routers: {p.routers}")

__all__ = ("plugins",)



import inspect
from inspect import Parameter
from typing import Any, Dict

from ..models import ProviderFn


def convert(value: str, type_: type) -> Any:
    """Convert string `value` to type `type_`."""
    if type_ is bool:
        return value.lower() in ("1", "t", "true", "y", "yes")
    return type_(value)


async def fetch_data(
    provider: ProviderFn, path_params: Dict[str, Any], query_params: Dict[str, str]
) -> Any:
    """Fetch data by using the data provider function.

    We first inspect the signature of the `provider` function to determine what
    parameters it needs. We then find values for these parameters from `path_params` &
    `query_params`. Finally we invoke `provider(**kwds)` and return the fetched data.

    Args:
        provider: The data provider function.
        path_params: `request.path_params`, arguments parsed from the route path.
        query_params: `request.query_params`, arguments parsed from the query string,
            but deduplicated.

    Returns:
        The data fetched by the `provider` function.

    Raises:
        NotImplementedError: do not use positional-only arguments in `provider`.
    """
    kwds = {}
    for parameter in inspect.signature(provider).parameters.values():
        # ignore `*args` & `**kwds` arguments
        if parameter.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
            continue
        parameter_value = look_up_parameter_value(parameter, path_params, query_params)
        if parameter_value is None:
            continue
        # handle positional-or-keyword & keyword-only arguments
        elif parameter.kind in (
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.KEYWORD_ONLY,
        ):
            kwds[parameter.name] = parameter_value
        # do not use positional-only arguments
        # TODO: document this behavior
        else:
            raise NotImplementedError()  # pragma: no cover
    return await provider(**kwds)


def look_up_parameter_value(
    parameter: Parameter, path_params: Dict[str, Any], query_params: Dict[str, str]
) -> Any:
    """Look up value for a specific parameter from `path_params` & `query_params`.

    Args:
        parameter: A parameter of some data provider function (`ProviderFn`).
        path_params: `request.path_params`, arguments parsed from the route.
        query_params: `request.query_params`, arguments parsed from the query string,
            but deduplicated.

    Returns:
        The value for the parameter, `None` if not found.
    """
    value = None
    # look for `parameter.name` in `path_params`
    if parameter.name in path_params:
        value = path_params[parameter.name]
        if isinstance(value, str):
            value = convert(value, parameter.annotation)
    # looking for `parameter.name` in `query_params`
    elif parameter.name in query_params:
        value = convert(query_params[parameter.name], parameter.annotation)
    return value



import asyncio
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional

from pydantic import BaseModel, validator
from pydantic.class_validators import root_validator


if TYPE_CHECKING:
    EmailStr = str
    HttpUrl = str
else:
    from pydantic import EmailStr, HttpUrl


ProviderFn = Callable[..., Awaitable[Dict[str, Any]]]


class PluginModelError(ValueError):
    """Exception for `Plugin` model validation error."""

    empty_router = "plugin must include at least one router."
    provider_not_async = "provider functions must be asynchronous."
    unexpected_plugin_name = 'plugin names must start with "rsserpent-plugin-".'
    unexpected_router_path = "all path in `routers` must starts with `prefix`."


class Persona(BaseModel):
    """Data model for plugin authors' personal information."""

    name: str
    link: HttpUrl
    email: EmailStr


class Plugin(BaseModel):
    """Data model for a RSSerpent plugin."""

    name: str
    author: Persona
    repository: HttpUrl
    prefix: str
    routers: Dict[str, ProviderFn]

    @root_validator
    def validate(  # type: ignore[override]
        cls, values: Dict[str, Any]  # noqa: N805
    ) -> Dict[str, Any]:
        """Ensure all paths in `routers` starts with `prefix`."""
        prefix: Optional[str] = values.get("prefix")
        routers: Optional[Dict[str, ProviderFn]] = values.get("routers")
        assert prefix is not None and routers is not None
        for path in routers:
            if not path.startswith(prefix):
                raise PluginModelError(PluginModelError.unexpected_router_path)
        return values

    @validator("name")
    def validate_name(cls, name: str) -> str:  # noqa: N805
        r"""Ensure any plugin name starts with `"rsserpent-plugin-"`."""
        if not name.startswith("rsserpent-plugin-"):
            raise PluginModelError(PluginModelError.unexpected_plugin_name)
        return name

    @validator("routers")
    def validate_routers(
        cls, routers: Dict[str, ProviderFn]  # noqa: N805
    ) -> Dict[str, ProviderFn]:
        """Ensure `routers` is not empty & all provider functions are async."""
        if len(routers) < 1:
            raise PluginModelError(PluginModelError.empty_router)
        for provider in routers.values():
            if not asyncio.iscoroutinefunction(provider):
                raise PluginModelError(PluginModelError.provider_not_async)
        return routers



from .cache import cached
from .http import Browser, HTTPClient
from .provider import fetch_data
from .ratelimit import ratelimit


__all__ = ("Browser", "HTTPClient", "cached", "fetch_data", "ratelimit")



from typing import Any, Dict


path = "/_/example/{n:int}"


async def provider(n: int) -> Dict[str, Any]:
    """Define a example data provider function with arguments."""
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [
            {"title": f"Example Title {i}", "description": f"Example Description {i}"}
            for i in range(1, n + 1)
        ],
    }
